/*
 * livepatch_bsc1192991
 *
 * Fix for CVE-2021-4001, bsc#1192991
 *
 *  Upstream commit:
 *  353050be4c19 ("bpf: Fix toctou on read-only map's constant scalar tracking")
 *
 *  SLE12-SP3 commit:
 *  not affected
 *
 *  SLE12-SP4, SLE12-SP5, SLE15 and SLE15-SP1 commit:
 *  not affected
 *
 *  SLE15-SP2 and -SP3 commit:
 *  b9b6f650fda1b6a8024f492c5b1e2c3c5e3c1001
 *
 *
 *  Copyright (c) 2022 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include "../klp_syscalls.h"

/* klp-ccp: from kernel/bpf/syscall.c */
#include <linux/bpf.h>

/* klp-ccp: from include/linux/bpf.h */

static struct idr (*klpe_btf_idr);
static spinlock_t (*klpe_btf_idr_lock);

static int (*klpe_bpf_obj_name_cpy)(char *dst, const char *src, unsigned int size);

static const struct bpf_map_ops (*klpe_bpf_map_offload_ops);

#ifdef CONFIG_BPF_SYSCALL
static int (*klpe_bpf_prog_active);
static struct mutex (*klpe_bpf_stats_enabled_mutex);

static inline void klpr_bpf_disable_instrumentation(void)
{
	migrate_disable();
	if (IS_ENABLED(CONFIG_PREEMPT_RT))
		this_cpu_generic_to_op((*klpe_bpf_prog_active), 1, +=);
	else
		raw_cpu_generic_to_op((*klpe_bpf_prog_active), 1, +=);
}

static inline void klpr_bpf_enable_instrumentation(void)
{
	if (IS_ENABLED(CONFIG_PREEMPT_RT))
		this_cpu_generic_to_op((*klpe_bpf_prog_active), -1, +=);
	else
		raw_cpu_generic_to_op((*klpe_bpf_prog_active), -1, +=);
	migrate_enable();
}

static struct bpf_prog *(*klpe_bpf_prog_get)(u32 ufd);

static struct bpf_map *(*klpe___bpf_map_get)(struct fd f);

static void (*klpe_bpf_map_put_with_uref)(struct bpf_map *map);

static void (*klpe_bpf_map_charge_finish)(struct bpf_map_memory *mem);
static void (*klpe_bpf_map_charge_move)(struct bpf_map_memory *dst,
			 struct bpf_map_memory *src);

static int (*klpe_sysctl_unprivileged_bpf_disabled);

static int (*klpe_bpf_map_new_fd)(struct bpf_map *map, int flags);
static int (*klpe_bpf_prog_new_fd)(struct bpf_prog *prog);

static void (*klpe_bpf_link_put)(struct bpf_link *link);
static int (*klpe_bpf_link_new_fd)(struct bpf_link *link);

static struct bpf_link *(*klpe_bpf_link_get_from_fd)(u32 ufd);

static int (*klpe_bpf_obj_pin_user)(u32 ufd, const char __user *pathname);
static int (*klpe_bpf_obj_get_user)(const char __user *pathname, int flags);

static int (*klpe_bpf_iter_link_attach)(const union bpf_attr *attr, struct bpf_prog *prog);
static int (*klpe_bpf_iter_new_fd)(struct bpf_link *link);

static int (*klpe_bpf_percpu_hash_update)(struct bpf_map *map, void *key, void *value,
			   u64 flags);
static int (*klpe_bpf_percpu_array_update)(struct bpf_map *map, void *key, void *value,
			    u64 flags);

static int (*klpe_bpf_fd_array_map_update_elem)(struct bpf_map *map, struct file *map_file,
				 void *key, void *value, u64 map_flags);

static int (*klpe_bpf_fd_htab_map_update_elem)(struct bpf_map *map, struct file *map_file,
				void *key, void *value, u64 map_flags);

static int (*klpe_bpf_get_file_flag)(int flags);
static int (*klpe_bpf_check_uarg_tail_zero)(void __user *uaddr, size_t expected_size,
			     size_t actual_size);

static struct bpf_prog *(*klpe_bpf_prog_by_id)(u32 id);

#else /* !CONFIG_BPF_SYSCALL */
#error "klp-ccp: non-taken branch"
#endif /* CONFIG_BPF_SYSCALL */

static int (*klpe_bpf_map_offload_update_elem)(struct bpf_map *map,
				void *key, void *value, u64 flags);
static int (*klpe_bpf_map_offload_delete_elem)(struct bpf_map *map, void *key);
static int (*klpe_bpf_map_offload_get_next_key)(struct bpf_map *map,
				 void *key, void *next_key);

#if defined(CONFIG_NET) && defined(CONFIG_BPF_SYSCALL)

static inline bool klpr_bpf_map_is_dev_bound(struct bpf_map *map)
{
	return unlikely(map->ops == &(*klpe_bpf_map_offload_ops));
}

#else
#error "klp-ccp: non-taken branch"
#endif /* CONFIG_NET && CONFIG_BPF_SYSCALL */

#if defined(CONFIG_BPF_STREAM_PARSER)

static int (*klpe_sock_map_get_from_fd)(const union bpf_attr *attr, struct bpf_prog *prog);
static int (*klpe_sock_map_prog_detach)(const union bpf_attr *attr, enum bpf_prog_type ptype);
static int (*klpe_sock_map_update_elem_sys)(struct bpf_map *map, void *key, void *value, u64 flags);

#else
#error "klp-ccp: non-taken branch"
#endif /* CONFIG_BPF_STREAM_PARSER */

#if defined(CONFIG_INET) && defined(CONFIG_BPF_SYSCALL)

static int (*klpe_bpf_fd_reuseport_array_update_elem)(struct bpf_map *map, void *key,
				       void *value, u64 map_flags);
#else
#error "klp-ccp: non-taken branch"
#endif /* defined(CONFIG_INET) && defined(CONFIG_BPF_SYSCALL) */

/* klp-ccp: from kernel/bpf/syscall.c */
#include <linux/bpf_trace.h>

/* klp-ccp: from include/linux/bpf-netns.h */
#ifdef CONFIG_NET
static int (*klpe_netns_bpf_prog_query)(const union bpf_attr *attr,
			 union bpf_attr __user *uattr);
static int (*klpe_netns_bpf_prog_attach)(const union bpf_attr *attr,
			  struct bpf_prog *prog);
static int (*klpe_netns_bpf_prog_detach)(const union bpf_attr *attr, enum bpf_prog_type ptype);
static int (*klpe_netns_bpf_link_create)(const union bpf_attr *attr,
			  struct bpf_prog *prog);
#else
#error "klp-ccp: non-taken branch"
#endif

/* klp-ccp: from include/linux/bpf-cgroup.h */
#ifdef CONFIG_CGROUP_BPF

static int (*klpe_bpf_percpu_cgroup_storage_update)(struct bpf_map *map, void *key,
				     void *value, u64 flags);

static int (*klpe_cgroup_bpf_prog_attach)(const union bpf_attr *attr,
			   enum bpf_prog_type ptype, struct bpf_prog *prog);
static int (*klpe_cgroup_bpf_prog_detach)(const union bpf_attr *attr,
			   enum bpf_prog_type ptype);
static int (*klpe_cgroup_bpf_link_attach)(const union bpf_attr *attr, struct bpf_prog *prog);
static int (*klpe_cgroup_bpf_prog_query)(const union bpf_attr *attr,
			  union bpf_attr __user *uattr);
#else
#error "klp-ccp: non-taken branch"
#endif /* CONFIG_CGROUP_BPF */

/* klp-ccp: from include/linux/security.h */
#ifdef CONFIG_BPF_SYSCALL

#ifdef CONFIG_SECURITY
static int (*klpe_security_bpf)(int cmd, union bpf_attr *attr, unsigned int size);

static int (*klpe_security_bpf_map_alloc)(struct bpf_map *map);
static void (*klpe_security_bpf_map_free)(struct bpf_map *map);

#else
#error "klp-ccp: non-taken branch"
#endif /* CONFIG_SECURITY */
#else
#error "klp-ccp: a preceeding branch should have been taken"
#endif /* CONFIG_BPF_SYSCALL */

/* klp-ccp: from include/linux/netdevice.h */
static int (*klpe_bpf_xdp_link_attach)(const union bpf_attr *attr, struct bpf_prog *prog);


/* klp-ccp: from kernel/bpf/syscall.c */
#include <linux/bpf_lirc.h>

/* klp-ccp: from include/linux/btf.h */
static void (*klpe_btf_put)(struct btf *btf);
static int (*klpe_btf_new_fd)(const union bpf_attr *attr);
static struct btf *(*klpe_btf_get_by_fd)(int fd);

static const struct btf_type *(*klpe_btf_type_id_size)(const struct btf *btf,
					u32 *type_id,
					u32 *ret_size);

static int (*klpe_btf_get_fd_by_id)(u32 id);

static int (*klpe_btf_find_spin_lock)(const struct btf *btf, const struct btf_type *t);

#ifdef CONFIG_BPF_SYSCALL
static const struct btf_type *(*klpe_btf_type_by_id)(const struct btf *btf, u32 type_id);

#else
#error "klp-ccp: non-taken branch"
#endif

/* klp-ccp: from include/linux/perf_event.h */
#ifdef CONFIG_PERF_EVENTS

static const struct perf_event *(*klpe_perf_get_event)(struct file *file);

#else /* !CONFIG_PERF_EVENTS: */
#error "klp-ccp: non-taken branch"
#endif

/* klp-ccp: from include/linux/trace_events.h */
#ifdef CONFIG_BPF_EVENTS

static int (*klpe_bpf_get_perf_event_info)(const struct perf_event *event, u32 *prog_id,
			    u32 *fd_type, const char **buf,
			    u64 *probe_offset, u64 *probe_addr);
#else
#error "klp-ccp: non-taken branch"
#endif

/* klp-ccp: from kernel/bpf/syscall.c */
#include <linux/slab.h>
#include <linux/sched/signal.h>
#include <linux/vmalloc.h>
#include <linux/mmzone.h>
#include <linux/anon_inodes.h>
#include <linux/fdtable.h>

/* klp-ccp: from include/linux/fdtable.h */
static struct files_struct *(*klpe_get_files_struct)(struct task_struct *);
static void (*klpe_put_files_struct)(struct files_struct *fs);

/* klp-ccp: from kernel/bpf/syscall.c */
#include <linux/file.h>
#include <linux/fs.h>
#include <linux/filter.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/idr.h>
#include <linux/cred.h>
#include <linux/timekeeping.h>
#include <linux/nospec.h>
#include <uapi/linux/btf.h>
#include <linux/poll.h>
#include <linux/bpf-netns.h>

#define IS_FD_ARRAY(map) ((map)->map_type == BPF_MAP_TYPE_PERF_EVENT_ARRAY || \
			  (map)->map_type == BPF_MAP_TYPE_CGROUP_ARRAY || \
			  (map)->map_type == BPF_MAP_TYPE_ARRAY_OF_MAPS)
#define IS_FD_PROG_ARRAY(map) ((map)->map_type == BPF_MAP_TYPE_PROG_ARRAY)
#define IS_FD_HASH(map) ((map)->map_type == BPF_MAP_TYPE_HASH_OF_MAPS)
#define IS_FD_MAP(map) (IS_FD_ARRAY(map) || IS_FD_PROG_ARRAY(map) || \
			IS_FD_HASH(map))

#define BPF_OBJ_FLAG_MASK   (BPF_F_RDONLY | BPF_F_WRONLY)

static struct idr (*klpe_prog_idr);
static spinlock_t (*klpe_prog_idr_lock);
static struct idr (*klpe_map_idr);
static spinlock_t (*klpe_map_idr_lock);
static struct idr (*klpe_link_idr);
static spinlock_t (*klpe_link_idr_lock);

static const struct bpf_map_ops * const (*klpe_bpf_map_types)[28]

/* klp-ccp: from include/linux/bpf_types.h */
#ifdef CONFIG_NET

#ifdef CONFIG_CGROUP_BPF

#else
#error "klp-ccp: a preceeding branch should have been taken"
#endif

#else
#error "klp-ccp: a preceeding branch should have been taken"
#endif
#ifdef CONFIG_BPF_EVENTS

#else
#error "klp-ccp: a preceeding branch should have been taken"
#endif
#ifdef CONFIG_CGROUP_BPF

#else
#error "klp-ccp: a preceeding branch should have been taken"
#endif
#ifdef CONFIG_BPF_LIRC_MODE2
#error "klp-ccp: non-taken branch"
#endif
#ifdef CONFIG_INET

#else
#error "klp-ccp: a preceeding branch should have been taken"
#endif
#if defined(CONFIG_BPF_JIT)

#ifdef CONFIG_BPF_LSM

#else
#error "klp-ccp: a preceeding branch should have been taken"
#endif /* CONFIG_BPF_LSM */
#else
#error "klp-ccp: a preceeding branch should have been taken"
#endif

#ifdef CONFIG_CGROUPS

#else
#error "klp-ccp: a preceeding branch should have been taken"
#endif
#ifdef CONFIG_CGROUP_BPF

#else
#error "klp-ccp: a preceeding branch should have been taken"
#endif

#ifdef CONFIG_PERF_EVENTS

#else
#error "klp-ccp: a preceeding branch should have been taken"
#endif

#ifdef CONFIG_NET

#if defined(CONFIG_BPF_STREAM_PARSER)

#else
#error "klp-ccp: a preceeding branch should have been taken"
#endif

#if defined(CONFIG_XDP_SOCKETS)

#else
#error "klp-ccp: a preceeding branch should have been taken"
#endif
#ifdef CONFIG_INET

#else
#error "klp-ccp: a preceeding branch should have been taken"
#endif
#else
#error "klp-ccp: a preceeding branch should have been taken"
#endif

#if defined(CONFIG_BPF_JIT)

#else
#error "klp-ccp: a preceeding branch should have been taken"
#endif

#ifdef CONFIG_CGROUP_BPF

#else
#error "klp-ccp: a preceeding branch should have been taken"
#endif

#ifdef CONFIG_NET

#else
#error "klp-ccp: a preceeding branch should have been taken"
#endif
/* klp-ccp: from kernel/bpf/syscall.c */
;

static struct bpf_map *klpr_find_and_alloc_map(union bpf_attr *attr)
{
	const struct bpf_map_ops *ops;
	u32 type = attr->map_type;
	struct bpf_map *map;
	int err;

	if (type >= ARRAY_SIZE((*klpe_bpf_map_types)))
		return ERR_PTR(-EINVAL);
	type = array_index_nospec(type, ARRAY_SIZE((*klpe_bpf_map_types)));
	ops = (*klpe_bpf_map_types)[type];
	if (!ops)
		return ERR_PTR(-EINVAL);

	if (ops->map_alloc_check) {
		err = ops->map_alloc_check(attr);
		if (err)
			return ERR_PTR(err);
	}
	if (attr->map_ifindex)
		ops = &(*klpe_bpf_map_offload_ops);
	map = ops->map_alloc(attr);
	if (IS_ERR(map))
		return map;
	map->ops = ops;
	map->map_type = type;
	return map;
}

static u32 bpf_map_value_size(struct bpf_map *map)
{
	if (map->map_type == BPF_MAP_TYPE_PERCPU_HASH ||
	    map->map_type == BPF_MAP_TYPE_LRU_PERCPU_HASH ||
	    map->map_type == BPF_MAP_TYPE_PERCPU_ARRAY ||
	    map->map_type == BPF_MAP_TYPE_PERCPU_CGROUP_STORAGE)
		return round_up(map->value_size, 8) * num_possible_cpus();
	else if (IS_FD_MAP(map))
		return sizeof(u32);
	else
		return  map->value_size;
}

static void maybe_wait_bpf_programs(struct bpf_map *map)
{
	/* Wait for any running BPF programs to complete so that
	 * userspace, when we return to it, knows that all programs
	 * that could be running use the new map value.
	 */
	if (map->map_type == BPF_MAP_TYPE_HASH_OF_MAPS ||
	    map->map_type == BPF_MAP_TYPE_ARRAY_OF_MAPS)
		synchronize_rcu();
}

static int klpr_bpf_map_update_value(struct bpf_map *map, struct fd f, void *key,
				void *value, __u64 flags)
{
	int err;

	/* Need to create a kthread, thus must support schedule */
	if (klpr_bpf_map_is_dev_bound(map)) {
		return (*klpe_bpf_map_offload_update_elem)(map, key, value, flags);
	} else if (map->map_type == BPF_MAP_TYPE_CPUMAP ||
		   map->map_type == BPF_MAP_TYPE_STRUCT_OPS) {
		return map->ops->map_update_elem(map, key, value, flags);
	} else if (map->map_type == BPF_MAP_TYPE_SOCKHASH ||
		   map->map_type == BPF_MAP_TYPE_SOCKMAP) {
		return (*klpe_sock_map_update_elem_sys)(map, key, value, flags);
	} else if (IS_FD_PROG_ARRAY(map)) {
		return (*klpe_bpf_fd_array_map_update_elem)(map, f.file, key, value,
						    flags);
	}

	klpr_bpf_disable_instrumentation();
	if (map->map_type == BPF_MAP_TYPE_PERCPU_HASH ||
	    map->map_type == BPF_MAP_TYPE_LRU_PERCPU_HASH) {
		err = (*klpe_bpf_percpu_hash_update)(map, key, value, flags);
	} else if (map->map_type == BPF_MAP_TYPE_PERCPU_ARRAY) {
		err = (*klpe_bpf_percpu_array_update)(map, key, value, flags);
	} else if (map->map_type == BPF_MAP_TYPE_PERCPU_CGROUP_STORAGE) {
		err = (*klpe_bpf_percpu_cgroup_storage_update)(map, key, value,
						       flags);
	} else if (IS_FD_ARRAY(map)) {
		rcu_read_lock();
		err = (*klpe_bpf_fd_array_map_update_elem)(map, f.file, key, value,
						   flags);
		rcu_read_unlock();
	} else if (map->map_type == BPF_MAP_TYPE_HASH_OF_MAPS) {
		rcu_read_lock();
		err = (*klpe_bpf_fd_htab_map_update_elem)(map, f.file, key, value,
						  flags);
		rcu_read_unlock();
	} else if (map->map_type == BPF_MAP_TYPE_REUSEPORT_SOCKARRAY) {
		/* rcu_read_lock() is not needed */
		err = (*klpe_bpf_fd_reuseport_array_update_elem)(map, key, value,
							 flags);
	} else if (map->map_type == BPF_MAP_TYPE_QUEUE ||
		   map->map_type == BPF_MAP_TYPE_STACK) {
		err = map->ops->map_push_elem(map, value, flags);
	} else {
		rcu_read_lock();
		err = map->ops->map_update_elem(map, key, value, flags);
		rcu_read_unlock();
	}
	klpr_bpf_enable_instrumentation();
	maybe_wait_bpf_programs(map);

	return err;
}

static int (*klpe_bpf_map_copy_value)(struct bpf_map *map, void *key, void *value,
			      __u64 flags);

static int klpr_bpf_map_alloc_id(struct bpf_map *map)
{
	int id;

	idr_preload(GFP_KERNEL);
	spin_lock_bh(&(*klpe_map_idr_lock));
	id = idr_alloc_cyclic(&(*klpe_map_idr), map, 1, INT_MAX, GFP_ATOMIC);
	if (id > 0)
		map->id = id;
	spin_unlock_bh(&(*klpe_map_idr_lock));
	idr_preload_end();

	if (WARN_ON_ONCE(!id))
		return -ENOSPC;

	return id > 0 ? 0 : id;
}

static fmode_t map_get_sys_perms(struct bpf_map *map, struct fd f)
{
	fmode_t mode = f.file->f_mode;

	/* Our file permissions may have been overridden by global
	 * map permissions facing syscall side.
	 */
	if (READ_ONCE(map->frozen))
		mode &= ~FMODE_CAN_WRITE;
	return mode;
}

#define CHECK_ATTR(CMD) \
	memchr_inv((void *) &attr->CMD##_LAST_FIELD + \
		   sizeof(attr->CMD##_LAST_FIELD), 0, \
		   sizeof(*attr) - \
		   offsetof(union bpf_attr, CMD##_LAST_FIELD) - \
		   sizeof(attr->CMD##_LAST_FIELD)) != NULL

static int klpr_map_check_btf(struct bpf_map *map, const struct btf *btf,
			 u32 btf_key_id, u32 btf_value_id)
{
	const struct btf_type *key_type, *value_type;
	u32 key_size, value_size;
	int ret = 0;

	/* Some maps allow key to be unspecified. */
	if (btf_key_id) {
		key_type = (*klpe_btf_type_id_size)(btf, &btf_key_id, &key_size);
		if (!key_type || key_size != map->key_size)
			return -EINVAL;
	} else {
		key_type = (*klpe_btf_type_by_id)(btf, 0);
		if (!map->ops->map_check_btf)
			return -EINVAL;
	}

	value_type = (*klpe_btf_type_id_size)(btf, &btf_value_id, &value_size);
	if (!value_type || value_size != map->value_size)
		return -EINVAL;

	map->spin_lock_off = (*klpe_btf_find_spin_lock)(btf, value_type);

	if (map_value_has_spin_lock(map)) {
		if (map->map_flags & BPF_F_RDONLY_PROG)
			return -EACCES;
		if (map->map_type != BPF_MAP_TYPE_HASH &&
		    map->map_type != BPF_MAP_TYPE_ARRAY &&
		    map->map_type != BPF_MAP_TYPE_CGROUP_STORAGE &&
		    map->map_type != BPF_MAP_TYPE_SK_STORAGE)
			return -ENOTSUPP;
		if (map->spin_lock_off + sizeof(struct bpf_spin_lock) >
		    map->value_size) {
			WARN_ONCE(1,
				  "verifier bug spin_lock_off %d value_size %d\n",
				  map->spin_lock_off, map->value_size);
			return -EFAULT;
		}
	}

	if (map->ops->map_check_btf)
		ret = map->ops->map_check_btf(map, btf, key_type, value_type);

	return ret;
}

#define BPF_MAP_CREATE_LAST_FIELD btf_vmlinux_value_type_id

static int klpr_map_create(union bpf_attr *attr)
{
	int numa_node = bpf_map_attr_numa_node(attr);
	struct bpf_map_memory mem;
	struct bpf_map *map;
	int f_flags;
	int err;

	err = CHECK_ATTR(BPF_MAP_CREATE);
	if (err)
		return -EINVAL;

	if (attr->btf_vmlinux_value_type_id) {
		if (attr->map_type != BPF_MAP_TYPE_STRUCT_OPS ||
		    attr->btf_key_type_id || attr->btf_value_type_id)
			return -EINVAL;
	} else if (attr->btf_key_type_id && !attr->btf_value_type_id) {
		return -EINVAL;
	}

	f_flags = (*klpe_bpf_get_file_flag)(attr->map_flags);
	if (f_flags < 0)
		return f_flags;

	if (numa_node != NUMA_NO_NODE &&
	    ((unsigned int)numa_node >= nr_node_ids ||
	     !node_online(numa_node)))
		return -EINVAL;

	/* find map type and init map: hashtable vs rbtree vs bloom vs ... */
	map = klpr_find_and_alloc_map(attr);
	if (IS_ERR(map))
		return PTR_ERR(map);

	err = (*klpe_bpf_obj_name_cpy)(map->name, attr->map_name,
			       sizeof(attr->map_name));
	if (err < 0)
		goto free_map;

	atomic64_set(&map->refcnt, 1);
	atomic64_set(&map->usercnt, 1);
	mutex_init(&map->freeze_mutex);

	map->spin_lock_off = -EINVAL;
	if (attr->btf_key_type_id || attr->btf_value_type_id ||
	    /* Even the map's value is a kernel's struct,
	     * the bpf_prog.o must have BTF to begin with
	     * to figure out the corresponding kernel's
	     * counter part.  Thus, attr->btf_fd has
	     * to be valid also.
	     */
	    attr->btf_vmlinux_value_type_id) {
		struct btf *btf;

		btf = (*klpe_btf_get_by_fd)(attr->btf_fd);
		if (IS_ERR(btf)) {
			err = PTR_ERR(btf);
			goto free_map;
		}
		map->btf = btf;

		if (attr->btf_value_type_id) {
			err = klpr_map_check_btf(map, btf, attr->btf_key_type_id,
					    attr->btf_value_type_id);
			if (err)
				goto free_map;
		}

		map->btf_key_type_id = attr->btf_key_type_id;
		map->btf_value_type_id = attr->btf_value_type_id;
		map->btf_vmlinux_value_type_id =
			attr->btf_vmlinux_value_type_id;
	}

	err = (*klpe_security_bpf_map_alloc)(map);
	if (err)
		goto free_map;

	err = klpr_bpf_map_alloc_id(map);
	if (err)
		goto free_map_sec;

	err = (*klpe_bpf_map_new_fd)(map, f_flags);
	if (err < 0) {
		/* failed to allocate fd.
		 * bpf_map_put_with_uref() is needed because the above
		 * bpf_map_alloc_id() has published the map
		 * to the userspace and the userspace may
		 * have refcnt-ed it through BPF_MAP_GET_FD_BY_ID.
		 */
		(*klpe_bpf_map_put_with_uref)(map);
		return err;
	}

	return err;

free_map_sec:
	(*klpe_security_bpf_map_free)(map);
free_map:
	(*klpe_btf_put)(map->btf);
	(*klpe_bpf_map_charge_move)(&mem, &map->memory);
	map->ops->map_free(map);
	(*klpe_bpf_map_charge_finish)(&mem);
	return err;
}

static struct bpf_map *(*klpe___bpf_map_inc_not_zero)(struct bpf_map *map, bool uref);

static void *__bpf_copy_key(void __user *ukey, u64 key_size)
{
	if (key_size)
		return memdup_user(ukey, key_size);

	if (ukey)
		return ERR_PTR(-EINVAL);

	return NULL;
}

#define BPF_MAP_LOOKUP_ELEM_LAST_FIELD flags

static int klpr_map_lookup_elem(union bpf_attr *attr)
{
	void __user *ukey = u64_to_user_ptr(attr->key);
	void __user *uvalue = u64_to_user_ptr(attr->value);
	int ufd = attr->map_fd;
	struct bpf_map *map;
	void *key, *value;
	u32 value_size;
	struct fd f;
	int err;

	if (CHECK_ATTR(BPF_MAP_LOOKUP_ELEM))
		return -EINVAL;

	if (attr->flags & ~BPF_F_LOCK)
		return -EINVAL;

	f = fdget(ufd);
	map = (*klpe___bpf_map_get)(f);
	if (IS_ERR(map))
		return PTR_ERR(map);
	if (!(map_get_sys_perms(map, f) & FMODE_CAN_READ)) {
		err = -EPERM;
		goto err_put;
	}

	if ((attr->flags & BPF_F_LOCK) &&
	    !map_value_has_spin_lock(map)) {
		err = -EINVAL;
		goto err_put;
	}

	key = __bpf_copy_key(ukey, map->key_size);
	if (IS_ERR(key)) {
		err = PTR_ERR(key);
		goto err_put;
	}

	value_size = bpf_map_value_size(map);

	err = -ENOMEM;
	value = kmalloc(value_size, GFP_USER | __GFP_NOWARN);
	if (!value)
		goto free_key;

	err = (*klpe_bpf_map_copy_value)(map, key, value, attr->flags);
	if (err)
		goto free_value;

	err = -EFAULT;
	if (copy_to_user(uvalue, value, value_size) != 0)
		goto free_value;

	err = 0;

free_value:
	kfree(value);
free_key:
	kfree(key);
err_put:
	fdput(f);
	return err;
}

#define BPF_MAP_UPDATE_ELEM_LAST_FIELD flags

static int klpp_map_update_elem(union bpf_attr *attr)
{
	void __user *ukey = u64_to_user_ptr(attr->key);
	void __user *uvalue = u64_to_user_ptr(attr->value);
	int ufd = attr->map_fd;
	struct bpf_map *map;
	void *key, *value;
	u32 value_size;
	struct fd f;
	int err;

	if (CHECK_ATTR(BPF_MAP_UPDATE_ELEM))
		return -EINVAL;

	f = fdget(ufd);
	map = (*klpe___bpf_map_get)(f);
	if (IS_ERR(map))
		return PTR_ERR(map);
	/*
	 * Fix CVE-2021-4001
	 *  +2 lines
	 */
	if (map->map_flags & BPF_F_RDONLY_PROG)
		mutex_lock(&map->freeze_mutex);
	if (!(map_get_sys_perms(map, f) & FMODE_CAN_WRITE)) {
		/*
		 * Fix CVE-2021-4001
		 *  +2 lines
		 */
		if (map->map_flags & BPF_F_RDONLY_PROG)
			mutex_unlock(&map->freeze_mutex);
		err = -EPERM;
		goto err_put;
	}
	/*
	 * Fix CVE-2021-4001
	 *  +4 lines
	 */
	if (map->map_flags & BPF_F_RDONLY_PROG) {
		map->writecnt++;
		mutex_unlock(&map->freeze_mutex);
	}

	if ((attr->flags & BPF_F_LOCK) &&
	    !map_value_has_spin_lock(map)) {
		err = -EINVAL;
		/*
		 * Fix CVE-2021-4001
		 *  -1 line, +1 line
		 */
		goto dec_writecnt;
	}

	key = __bpf_copy_key(ukey, map->key_size);
	if (IS_ERR(key)) {
		err = PTR_ERR(key);
		/*
		 * Fix CVE-2021-4001
		 *  -1 line, +1 line
		 */
		goto dec_writecnt;
	}

	if (map->map_type == BPF_MAP_TYPE_PERCPU_HASH ||
	    map->map_type == BPF_MAP_TYPE_LRU_PERCPU_HASH ||
	    map->map_type == BPF_MAP_TYPE_PERCPU_ARRAY ||
	    map->map_type == BPF_MAP_TYPE_PERCPU_CGROUP_STORAGE)
		value_size = round_up(map->value_size, 8) * num_possible_cpus();
	else
		value_size = map->value_size;

	err = -ENOMEM;
	value = kmalloc(value_size, GFP_USER | __GFP_NOWARN);
	if (!value)
		goto free_key;

	err = -EFAULT;
	if (copy_from_user(value, uvalue, value_size) != 0)
		goto free_value;

	err = klpr_bpf_map_update_value(map, f, key, value, attr->flags);

free_value:
	kfree(value);
free_key:
	kfree(key);
/*
 * Fix CVE-2021-4001
 *  +6 lines
 */
dec_writecnt:
	if (map->map_flags & BPF_F_RDONLY_PROG) {
		mutex_lock(&map->freeze_mutex);
		map->writecnt--;
		mutex_unlock(&map->freeze_mutex);
	}
err_put:
	fdput(f);
	return err;
}

#define BPF_MAP_DELETE_ELEM_LAST_FIELD key

static int klpp_map_delete_elem(union bpf_attr *attr)
{
	void __user *ukey = u64_to_user_ptr(attr->key);
	int ufd = attr->map_fd;
	struct bpf_map *map;
	struct fd f;
	void *key;
	int err;

	if (CHECK_ATTR(BPF_MAP_DELETE_ELEM))
		return -EINVAL;

	f = fdget(ufd);
	map = (*klpe___bpf_map_get)(f);
	if (IS_ERR(map))
		return PTR_ERR(map);
	/*
	 * Fix CVE-2021-4001
	 *  +2 lines
	 */
	if (map->map_flags & BPF_F_RDONLY_PROG)
		mutex_lock(&map->freeze_mutex);
	if (!(map_get_sys_perms(map, f) & FMODE_CAN_WRITE)) {
		/*
		 * Fix CVE-2021-4001
		 *  +2 lines
		 */
		if (map->map_flags & BPF_F_RDONLY_PROG)
			mutex_unlock(&map->freeze_mutex);
		err = -EPERM;
		goto err_put;
	}
	/*
	 * Fix CVE-2021-4001
	 *  +4 lines
	 */
	if (map->map_flags & BPF_F_RDONLY_PROG) {
		map->writecnt++;
		mutex_unlock(&map->freeze_mutex);
	}

	key = __bpf_copy_key(ukey, map->key_size);
	if (IS_ERR(key)) {
		err = PTR_ERR(key);
		/*
		 * Fix CVE-2021-4001
		 *  -1 line, +1 line
		 */
		goto dec_writecnt;
	}

	if (klpr_bpf_map_is_dev_bound(map)) {
		err = (*klpe_bpf_map_offload_delete_elem)(map, key);
		goto out;
	} else if (IS_FD_PROG_ARRAY(map) ||
		   map->map_type == BPF_MAP_TYPE_STRUCT_OPS) {
		/* These maps require sleepable context */
		err = map->ops->map_delete_elem(map, key);
		goto out;
	}

	klpr_bpf_disable_instrumentation();
	rcu_read_lock();
	err = map->ops->map_delete_elem(map, key);
	rcu_read_unlock();
	klpr_bpf_enable_instrumentation();
	maybe_wait_bpf_programs(map);
out:
	kfree(key);
/*
 * Fix CVE-2021-4001
 *  +6 lines
 */
dec_writecnt:
	if (map->map_flags & BPF_F_RDONLY_PROG) {
		mutex_lock(&map->freeze_mutex);
		map->writecnt--;
		mutex_unlock(&map->freeze_mutex);
	}
err_put:
	fdput(f);
	return err;
}

#define BPF_MAP_GET_NEXT_KEY_LAST_FIELD next_key

static int klpr_map_get_next_key(union bpf_attr *attr)
{
	void __user *ukey = u64_to_user_ptr(attr->key);
	void __user *unext_key = u64_to_user_ptr(attr->next_key);
	int ufd = attr->map_fd;
	struct bpf_map *map;
	void *key, *next_key;
	struct fd f;
	int err;

	if (CHECK_ATTR(BPF_MAP_GET_NEXT_KEY))
		return -EINVAL;

	f = fdget(ufd);
	map = (*klpe___bpf_map_get)(f);
	if (IS_ERR(map))
		return PTR_ERR(map);
	if (!(map_get_sys_perms(map, f) & FMODE_CAN_READ)) {
		err = -EPERM;
		goto err_put;
	}

	if (ukey) {
		key = __bpf_copy_key(ukey, map->key_size);
		if (IS_ERR(key)) {
			err = PTR_ERR(key);
			goto err_put;
		}
	} else {
		key = NULL;
	}

	err = -ENOMEM;
	next_key = kmalloc(map->key_size, GFP_USER);
	if (!next_key)
		goto free_key;

	if (klpr_bpf_map_is_dev_bound(map)) {
		err = (*klpe_bpf_map_offload_get_next_key)(map, key, next_key);
		goto out;
	}

	rcu_read_lock();
	err = map->ops->map_get_next_key(map, key, next_key);
	rcu_read_unlock();
out:
	if (err)
		goto free_next_key;

	err = -EFAULT;
	if (copy_to_user(unext_key, next_key, map->key_size) != 0)
		goto free_next_key;

	err = 0;

free_next_key:
	kfree(next_key);
free_key:
	kfree(key);
err_put:
	fdput(f);
	return err;
}

#define BPF_MAP_LOOKUP_AND_DELETE_ELEM_LAST_FIELD value

static int klpp_map_lookup_and_delete_elem(union bpf_attr *attr)
{
	void __user *ukey = u64_to_user_ptr(attr->key);
	void __user *uvalue = u64_to_user_ptr(attr->value);
	int ufd = attr->map_fd;
	struct bpf_map *map;
	void *key, *value;
	u32 value_size;
	struct fd f;
	int err;

	if (CHECK_ATTR(BPF_MAP_LOOKUP_AND_DELETE_ELEM))
		return -EINVAL;

	f = fdget(ufd);
	map = (*klpe___bpf_map_get)(f);
	if (IS_ERR(map))
		return PTR_ERR(map);
	/*
	 * Fix CVE-2021-4001
	 *  +2 lines
	 */
	if (map->map_flags & BPF_F_RDONLY_PROG)
		mutex_lock(&map->freeze_mutex);
	if (!(map_get_sys_perms(map, f) & FMODE_CAN_READ) ||
	    !(map_get_sys_perms(map, f) & FMODE_CAN_WRITE)) {
		/*
		 * Fix CVE-2021-4001
		 *  +2 lines
		 */
		if (map->map_flags & BPF_F_RDONLY_PROG)
			mutex_unlock(&map->freeze_mutex);
		err = -EPERM;
		goto err_put;
	}
	/*
	 * Fix CVE-2021-4001
	 *  +4 lines
	 */
	if (map->map_flags & BPF_F_RDONLY_PROG) {
		map->writecnt++;
		mutex_unlock(&map->freeze_mutex);
	}

	key = __bpf_copy_key(ukey, map->key_size);
	if (IS_ERR(key)) {
		err = PTR_ERR(key);
		/*
		 * Fix CVE-2021-4001
		 *  -1 line, +1 line
		 */
		goto dec_writecnt;
	}

	value_size = map->value_size;

	err = -ENOMEM;
	value = kmalloc(value_size, GFP_USER | __GFP_NOWARN);
	if (!value)
		goto free_key;

	if (map->map_type == BPF_MAP_TYPE_QUEUE ||
	    map->map_type == BPF_MAP_TYPE_STACK) {
		err = map->ops->map_pop_elem(map, value);
	} else {
		err = -ENOTSUPP;
	}

	if (err)
		goto free_value;

	if (copy_to_user(uvalue, value, value_size) != 0) {
		err = -EFAULT;
		goto free_value;
	}

	err = 0;

free_value:
	kfree(value);
free_key:
	kfree(key);
/*
 * Fix CVE-2021-4001
 *  +6 lines
 */
dec_writecnt:
	if (map->map_flags & BPF_F_RDONLY_PROG) {
		mutex_lock(&map->freeze_mutex);
		map->writecnt--;
		mutex_unlock(&map->freeze_mutex);
	}
err_put:
	fdput(f);
	return err;
}

#define BPF_MAP_FREEZE_LAST_FIELD map_fd

static int klpr_map_freeze(const union bpf_attr *attr)
{
	int err = 0, ufd = attr->map_fd;
	struct bpf_map *map;
	struct fd f;

	if (CHECK_ATTR(BPF_MAP_FREEZE))
		return -EINVAL;

	f = fdget(ufd);
	map = (*klpe___bpf_map_get)(f);
	if (IS_ERR(map))
		return PTR_ERR(map);

	if (map->map_type == BPF_MAP_TYPE_STRUCT_OPS) {
		fdput(f);
		return -ENOTSUPP;
	}

	mutex_lock(&map->freeze_mutex);

	if (map->writecnt) {
		err = -EBUSY;
		goto err_put;
	}
	if (READ_ONCE(map->frozen)) {
		err = -EBUSY;
		goto err_put;
	}
	if (!bpf_capable()) {
		err = -EPERM;
		goto err_put;
	}

	WRITE_ONCE(map->frozen, true);
err_put:
	mutex_unlock(&map->freeze_mutex);
	fdput(f);
	return err;
}

static int (*klpe_bpf_prog_load)(union bpf_attr *attr, union bpf_attr __user *uattr);

#define BPF_OBJ_LAST_FIELD file_flags

static int klpr_bpf_obj_pin(const union bpf_attr *attr)
{
	if (CHECK_ATTR(BPF_OBJ) || attr->file_flags != 0)
		return -EINVAL;

	return (*klpe_bpf_obj_pin_user)(attr->bpf_fd, u64_to_user_ptr(attr->pathname));
}

static int klpr_bpf_obj_get(const union bpf_attr *attr)
{
	if (CHECK_ATTR(BPF_OBJ) || attr->bpf_fd != 0 ||
	    attr->file_flags & ~BPF_OBJ_FLAG_MASK)
		return -EINVAL;

	return (*klpe_bpf_obj_get_user)(u64_to_user_ptr(attr->pathname),
				attr->file_flags);
}

static const struct file_operations (*klpe_bpf_link_fops);

struct bpf_raw_tp_link {
	struct bpf_link link;
	struct bpf_raw_event_map *btp;
};

static const struct bpf_link_ops (*klpe_bpf_raw_tp_link_lops);

static int (*klpe_bpf_raw_tracepoint_open)(const union bpf_attr *attr);

static int (*klpe_bpf_prog_attach_check_attach_type)(const struct bpf_prog *prog,
					     enum bpf_attach_type attach_type);

static enum bpf_prog_type
attach_type_to_prog_type(enum bpf_attach_type attach_type)
{
	switch (attach_type) {
	case BPF_CGROUP_INET_INGRESS:
	case BPF_CGROUP_INET_EGRESS:
		return BPF_PROG_TYPE_CGROUP_SKB;
		break;
	case BPF_CGROUP_INET_SOCK_CREATE:
	case BPF_CGROUP_INET_SOCK_RELEASE:
	case BPF_CGROUP_INET4_POST_BIND:
	case BPF_CGROUP_INET6_POST_BIND:
		return BPF_PROG_TYPE_CGROUP_SOCK;
	case BPF_CGROUP_INET4_BIND:
	case BPF_CGROUP_INET6_BIND:
	case BPF_CGROUP_INET4_CONNECT:
	case BPF_CGROUP_INET6_CONNECT:
	case BPF_CGROUP_INET4_GETPEERNAME:
	case BPF_CGROUP_INET6_GETPEERNAME:
	case BPF_CGROUP_INET4_GETSOCKNAME:
	case BPF_CGROUP_INET6_GETSOCKNAME:
	case BPF_CGROUP_UDP4_SENDMSG:
	case BPF_CGROUP_UDP6_SENDMSG:
	case BPF_CGROUP_UDP4_RECVMSG:
	case BPF_CGROUP_UDP6_RECVMSG:
		return BPF_PROG_TYPE_CGROUP_SOCK_ADDR;
	case BPF_CGROUP_SOCK_OPS:
		return BPF_PROG_TYPE_SOCK_OPS;
	case BPF_CGROUP_DEVICE:
		return BPF_PROG_TYPE_CGROUP_DEVICE;
	case BPF_SK_MSG_VERDICT:
		return BPF_PROG_TYPE_SK_MSG;
	case BPF_SK_SKB_STREAM_PARSER:
	case BPF_SK_SKB_STREAM_VERDICT:
		return BPF_PROG_TYPE_SK_SKB;
	case BPF_LIRC_MODE2:
		return BPF_PROG_TYPE_LIRC_MODE2;
	case BPF_FLOW_DISSECTOR:
		return BPF_PROG_TYPE_FLOW_DISSECTOR;
	case BPF_CGROUP_SYSCTL:
		return BPF_PROG_TYPE_CGROUP_SYSCTL;
	case BPF_CGROUP_GETSOCKOPT:
	case BPF_CGROUP_SETSOCKOPT:
		return BPF_PROG_TYPE_CGROUP_SOCKOPT;
	case BPF_TRACE_ITER:
		return BPF_PROG_TYPE_TRACING;
	case BPF_SK_LOOKUP:
		return BPF_PROG_TYPE_SK_LOOKUP;
	case BPF_XDP:
		return BPF_PROG_TYPE_XDP;
	default:
		return BPF_PROG_TYPE_UNSPEC;
	}
}

#define BPF_PROG_ATTACH_LAST_FIELD replace_bpf_fd

#define BPF_F_ATTACH_MASK \
	(BPF_F_ALLOW_OVERRIDE | BPF_F_ALLOW_MULTI | BPF_F_REPLACE)

static int klpr_bpf_prog_attach(const union bpf_attr *attr)
{
	enum bpf_prog_type ptype;
	struct bpf_prog *prog;
	int ret;

	if (CHECK_ATTR(BPF_PROG_ATTACH))
		return -EINVAL;

	if (attr->attach_flags & ~BPF_F_ATTACH_MASK)
		return -EINVAL;

	ptype = attach_type_to_prog_type(attr->attach_type);
	if (ptype == BPF_PROG_TYPE_UNSPEC)
		return -EINVAL;

	prog = bpf_prog_get_type(attr->attach_bpf_fd, ptype);
	if (IS_ERR(prog))
		return PTR_ERR(prog);

	if ((*klpe_bpf_prog_attach_check_attach_type)(prog, attr->attach_type)) {
		bpf_prog_put(prog);
		return -EINVAL;
	}

	switch (ptype) {
	case BPF_PROG_TYPE_SK_SKB:
	case BPF_PROG_TYPE_SK_MSG:
		ret = (*klpe_sock_map_get_from_fd)(attr, prog);
		break;
	case BPF_PROG_TYPE_LIRC_MODE2:
		ret = lirc_prog_attach(attr, prog);
		break;
	case BPF_PROG_TYPE_FLOW_DISSECTOR:
		ret = (*klpe_netns_bpf_prog_attach)(attr, prog);
		break;
	case BPF_PROG_TYPE_CGROUP_DEVICE:
	case BPF_PROG_TYPE_CGROUP_SKB:
	case BPF_PROG_TYPE_CGROUP_SOCK:
	case BPF_PROG_TYPE_CGROUP_SOCK_ADDR:
	case BPF_PROG_TYPE_CGROUP_SOCKOPT:
	case BPF_PROG_TYPE_CGROUP_SYSCTL:
	case BPF_PROG_TYPE_SOCK_OPS:
		ret = (*klpe_cgroup_bpf_prog_attach)(attr, ptype, prog);
		break;
	default:
		ret = -EINVAL;
	}

	if (ret)
		bpf_prog_put(prog);
	return ret;
}

#define BPF_PROG_DETACH_LAST_FIELD attach_type

static int klpr_bpf_prog_detach(const union bpf_attr *attr)
{
	enum bpf_prog_type ptype;

	if (CHECK_ATTR(BPF_PROG_DETACH))
		return -EINVAL;

	ptype = attach_type_to_prog_type(attr->attach_type);

	switch (ptype) {
	case BPF_PROG_TYPE_SK_MSG:
	case BPF_PROG_TYPE_SK_SKB:
		return (*klpe_sock_map_prog_detach)(attr, ptype);
	case BPF_PROG_TYPE_LIRC_MODE2:
		return lirc_prog_detach(attr);
	case BPF_PROG_TYPE_FLOW_DISSECTOR:
		return (*klpe_netns_bpf_prog_detach)(attr, ptype);
	case BPF_PROG_TYPE_CGROUP_DEVICE:
	case BPF_PROG_TYPE_CGROUP_SKB:
	case BPF_PROG_TYPE_CGROUP_SOCK:
	case BPF_PROG_TYPE_CGROUP_SOCK_ADDR:
	case BPF_PROG_TYPE_CGROUP_SOCKOPT:
	case BPF_PROG_TYPE_CGROUP_SYSCTL:
	case BPF_PROG_TYPE_SOCK_OPS:
		return (*klpe_cgroup_bpf_prog_detach)(attr, ptype);
	default:
		return -EINVAL;
	}
}

#define BPF_PROG_QUERY_LAST_FIELD query.prog_cnt

static int klpr_bpf_prog_query(const union bpf_attr *attr,
			  union bpf_attr __user *uattr)
{
	if (!capable(CAP_NET_ADMIN))
		return -EPERM;
	if (CHECK_ATTR(BPF_PROG_QUERY))
		return -EINVAL;
	if (attr->query.query_flags & ~BPF_F_QUERY_EFFECTIVE)
		return -EINVAL;

	switch (attr->query.attach_type) {
	case BPF_CGROUP_INET_INGRESS:
	case BPF_CGROUP_INET_EGRESS:
	case BPF_CGROUP_INET_SOCK_CREATE:
	case BPF_CGROUP_INET_SOCK_RELEASE:
	case BPF_CGROUP_INET4_BIND:
	case BPF_CGROUP_INET6_BIND:
	case BPF_CGROUP_INET4_POST_BIND:
	case BPF_CGROUP_INET6_POST_BIND:
	case BPF_CGROUP_INET4_CONNECT:
	case BPF_CGROUP_INET6_CONNECT:
	case BPF_CGROUP_INET4_GETPEERNAME:
	case BPF_CGROUP_INET6_GETPEERNAME:
	case BPF_CGROUP_INET4_GETSOCKNAME:
	case BPF_CGROUP_INET6_GETSOCKNAME:
	case BPF_CGROUP_UDP4_SENDMSG:
	case BPF_CGROUP_UDP6_SENDMSG:
	case BPF_CGROUP_UDP4_RECVMSG:
	case BPF_CGROUP_UDP6_RECVMSG:
	case BPF_CGROUP_SOCK_OPS:
	case BPF_CGROUP_DEVICE:
	case BPF_CGROUP_SYSCTL:
	case BPF_CGROUP_GETSOCKOPT:
	case BPF_CGROUP_SETSOCKOPT:
		return (*klpe_cgroup_bpf_prog_query)(attr, uattr);
	case BPF_LIRC_MODE2:
		return lirc_prog_query(attr, uattr);
	case BPF_FLOW_DISSECTOR:
	case BPF_SK_LOOKUP:
		return (*klpe_netns_bpf_prog_query)(attr, uattr);
	default:
		return -EINVAL;
	}
}

#define BPF_PROG_TEST_RUN_LAST_FIELD test.ctx_out

static int klpr_bpf_prog_test_run(const union bpf_attr *attr,
			     union bpf_attr __user *uattr)
{
	struct bpf_prog *prog;
	int ret = -ENOTSUPP;

	if (CHECK_ATTR(BPF_PROG_TEST_RUN))
		return -EINVAL;

	if ((attr->test.ctx_size_in && !attr->test.ctx_in) ||
	    (!attr->test.ctx_size_in && attr->test.ctx_in))
		return -EINVAL;

	if ((attr->test.ctx_size_out && !attr->test.ctx_out) ||
	    (!attr->test.ctx_size_out && attr->test.ctx_out))
		return -EINVAL;

	prog = (*klpe_bpf_prog_get)(attr->test.prog_fd);
	if (IS_ERR(prog))
		return PTR_ERR(prog);

	if (prog->aux->ops->test_run)
		ret = prog->aux->ops->test_run(prog, attr, uattr);

	bpf_prog_put(prog);
	return ret;
}

static int (*klpe_bpf_obj_get_next_id)(const union bpf_attr *attr,
			       union bpf_attr __user *uattr,
			       struct idr *idr,
			       spinlock_t *lock);

#define BPF_PROG_GET_FD_BY_ID_LAST_FIELD prog_id

static int klpr_bpf_prog_get_fd_by_id(const union bpf_attr *attr)
{
	struct bpf_prog *prog;
	u32 id = attr->prog_id;
	int fd;

	if (CHECK_ATTR(BPF_PROG_GET_FD_BY_ID))
		return -EINVAL;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	prog = (*klpe_bpf_prog_by_id)(id);
	if (IS_ERR(prog))
		return PTR_ERR(prog);

	fd = (*klpe_bpf_prog_new_fd)(prog);
	if (fd < 0)
		bpf_prog_put(prog);

	return fd;
}

#define BPF_MAP_GET_FD_BY_ID_LAST_FIELD open_flags

static int klpr_bpf_map_get_fd_by_id(const union bpf_attr *attr)
{
	struct bpf_map *map;
	u32 id = attr->map_id;
	int f_flags;
	int fd;

	if (CHECK_ATTR(BPF_MAP_GET_FD_BY_ID) ||
	    attr->open_flags & ~BPF_OBJ_FLAG_MASK)
		return -EINVAL;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	f_flags = (*klpe_bpf_get_file_flag)(attr->open_flags);
	if (f_flags < 0)
		return f_flags;

	spin_lock_bh(&(*klpe_map_idr_lock));
	map = idr_find(&(*klpe_map_idr), id);
	if (map)
		map = (*klpe___bpf_map_inc_not_zero)(map, true);
	else
		map = ERR_PTR(-ENOENT);
	spin_unlock_bh(&(*klpe_map_idr_lock));

	if (IS_ERR(map))
		return PTR_ERR(map);

	fd = (*klpe_bpf_map_new_fd)(map, f_flags);
	if (fd < 0)
		(*klpe_bpf_map_put_with_uref)(map);

	return fd;
}

static int (*klpe_bpf_obj_get_info_by_fd)(const union bpf_attr *attr,
				  union bpf_attr __user *uattr);

#define BPF_BTF_LOAD_LAST_FIELD btf_log_level

static int klpr_bpf_btf_load(const union bpf_attr *attr)
{
	if (CHECK_ATTR(BPF_BTF_LOAD))
		return -EINVAL;

	if (!bpf_capable())
		return -EPERM;

	return (*klpe_btf_new_fd)(attr);
}

#define BPF_BTF_GET_FD_BY_ID_LAST_FIELD btf_id

static int klpr_bpf_btf_get_fd_by_id(const union bpf_attr *attr)
{
	if (CHECK_ATTR(BPF_BTF_GET_FD_BY_ID))
		return -EINVAL;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	return (*klpe_btf_get_fd_by_id)(attr->btf_id);
}

static int bpf_task_fd_query_copy(const union bpf_attr *attr,
				    union bpf_attr __user *uattr,
				    u32 prog_id, u32 fd_type,
				    const char *buf, u64 probe_offset,
				    u64 probe_addr)
{
	char __user *ubuf = u64_to_user_ptr(attr->task_fd_query.buf);
	u32 len = buf ? strlen(buf) : 0, input_len;
	int err = 0;

	if (put_user(len, &uattr->task_fd_query.buf_len))
		return -EFAULT;
	input_len = attr->task_fd_query.buf_len;
	if (input_len && ubuf) {
		if (!len) {
			/* nothing to copy, just make ubuf NULL terminated */
			char zero = '\0';

			if (put_user(zero, ubuf))
				return -EFAULT;
		} else if (input_len >= len + 1) {
			/* ubuf can hold the string with NULL terminator */
			if (copy_to_user(ubuf, buf, len + 1))
				return -EFAULT;
		} else {
			/* ubuf cannot hold the string with NULL terminator,
			 * do a partial copy with NULL terminator.
			 */
			char zero = '\0';

			err = -ENOSPC;
			if (copy_to_user(ubuf, buf, input_len - 1))
				return -EFAULT;
			if (put_user(zero, ubuf + input_len - 1))
				return -EFAULT;
		}
	}

	if (put_user(prog_id, &uattr->task_fd_query.prog_id) ||
	    put_user(fd_type, &uattr->task_fd_query.fd_type) ||
	    put_user(probe_offset, &uattr->task_fd_query.probe_offset) ||
	    put_user(probe_addr, &uattr->task_fd_query.probe_addr))
		return -EFAULT;

	return err;
}

#define BPF_TASK_FD_QUERY_LAST_FIELD task_fd_query.probe_addr

static int klpr_bpf_task_fd_query(const union bpf_attr *attr,
			     union bpf_attr __user *uattr)
{
	pid_t pid = attr->task_fd_query.pid;
	u32 fd = attr->task_fd_query.fd;
	const struct perf_event *event;
	struct files_struct *files;
	struct task_struct *task;
	struct file *file;
	int err;

	if (CHECK_ATTR(BPF_TASK_FD_QUERY))
		return -EINVAL;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (attr->task_fd_query.flags != 0)
		return -EINVAL;

	task = get_pid_task(find_vpid(pid), PIDTYPE_PID);
	if (!task)
		return -ENOENT;

	files = (*klpe_get_files_struct)(task);
	put_task_struct(task);
	if (!files)
		return -ENOENT;

	err = 0;
	spin_lock(&files->file_lock);
	file = fcheck_files(files, fd);
	if (!file)
		err = -EBADF;
	else
		get_file(file);
	spin_unlock(&files->file_lock);
	(*klpe_put_files_struct)(files);

	if (err)
		goto out;

	if (file->f_op == &(*klpe_bpf_link_fops)) {
		struct bpf_link *link = file->private_data;

		if (link->ops == &(*klpe_bpf_raw_tp_link_lops)) {
			struct bpf_raw_tp_link *raw_tp =
				container_of(link, struct bpf_raw_tp_link, link);
			struct bpf_raw_event_map *btp = raw_tp->btp;

			err = bpf_task_fd_query_copy(attr, uattr,
						     raw_tp->link.prog->aux->id,
						     BPF_FD_TYPE_RAW_TRACEPOINT,
						     btp->tp->name, 0, 0);
			goto put_file;
		}
		goto out_not_supp;
	}

	event = (*klpe_perf_get_event)(file);
	if (!IS_ERR(event)) {
		u64 probe_offset, probe_addr;
		u32 prog_id, fd_type;
		const char *buf;

		err = (*klpe_bpf_get_perf_event_info)(event, &prog_id, &fd_type,
					      &buf, &probe_offset,
					      &probe_addr);
		if (!err)
			err = bpf_task_fd_query_copy(attr, uattr, prog_id,
						     fd_type, buf,
						     probe_offset,
						     probe_addr);
		goto put_file;
	}

out_not_supp:
	err = -ENOTSUPP;
put_file:
	fput(file);
out:
	return err;
}

#define BPF_MAP_BATCH_LAST_FIELD batch.flags

#define BPF_DO_BATCH(fn)			\
	do {					\
		if (!fn) {			\
			err = -ENOTSUPP;	\
			goto err_put;		\
		}				\
		err = fn(map, attr, uattr);	\
	} while (0)

static int klpp_bpf_map_do_batch(const union bpf_attr *attr,
			    union bpf_attr __user *uattr,
			    int cmd)
{
	struct bpf_map *map;
	int err, ufd;
	struct fd f;

	if (CHECK_ATTR(BPF_MAP_BATCH))
		return -EINVAL;

	ufd = attr->batch.map_fd;
	f = fdget(ufd);
	map = (*klpe___bpf_map_get)(f);
	if (IS_ERR(map))
		return PTR_ERR(map);

	if ((cmd == BPF_MAP_LOOKUP_BATCH ||
	     cmd == BPF_MAP_LOOKUP_AND_DELETE_BATCH) &&
	    !(map_get_sys_perms(map, f) & FMODE_CAN_READ)) {
		err = -EPERM;
		goto err_put;
	}

	/*
	 * Fix CVE-2021-4001
	 *  +2 lines
	 */
	if (cmd != BPF_MAP_LOOKUP_BATCH && map->map_flags & BPF_F_RDONLY_PROG)
		mutex_lock(&map->freeze_mutex);
	if (cmd != BPF_MAP_LOOKUP_BATCH &&
	    !(map_get_sys_perms(map, f) & FMODE_CAN_WRITE)) {
		/*
		 * Fix CVE-2021-4001
		 *  +2 lines
		 */
		if (map->map_flags & BPF_F_RDONLY_PROG)
			mutex_unlock(&map->freeze_mutex);
		err = -EPERM;
		goto err_put;
	}
	/*
	 * Fix CVE-2021-4001
	 *  +4 lines
	 */
	if (cmd != BPF_MAP_LOOKUP_BATCH && map->map_flags & BPF_F_RDONLY_PROG) {
		map->writecnt++;
		mutex_unlock(&map->freeze_mutex);
	}

/*
 * Fix CVE-2021-4001
 *  +1 line
 */
#define err_put dec_writecnt
	if (cmd == BPF_MAP_LOOKUP_BATCH)
		BPF_DO_BATCH(map->ops->map_lookup_batch);
	else if (cmd == BPF_MAP_LOOKUP_AND_DELETE_BATCH)
		BPF_DO_BATCH(map->ops->map_lookup_and_delete_batch);
	else if (cmd == BPF_MAP_UPDATE_BATCH)
		BPF_DO_BATCH(map->ops->map_update_batch);
	else
		BPF_DO_BATCH(map->ops->map_delete_batch);
/*
 * Fix CVE-2021-4001
 *  +1 line
 */
#undef err_put

/*
 * Fix CVE-2021-4001
 *  +6 lines
 */
dec_writecnt:
	if (cmd != BPF_MAP_LOOKUP_BATCH && map->map_flags & BPF_F_RDONLY_PROG) {
		mutex_lock(&map->freeze_mutex);
		map->writecnt--;
		mutex_unlock(&map->freeze_mutex);
	}
err_put:
	fdput(f);
	return err;
}

static int klpr_tracing_bpf_link_attach(const union bpf_attr *attr, struct bpf_prog *prog)
{
	if (attr->link_create.attach_type == BPF_TRACE_ITER &&
	    prog->expected_attach_type == BPF_TRACE_ITER)
		return (*klpe_bpf_iter_link_attach)(attr, prog);

	return -EINVAL;
}

#define BPF_LINK_CREATE_LAST_FIELD link_create.iter_info_len
static int klpr_link_create(union bpf_attr *attr)
{
	enum bpf_prog_type ptype;
	struct bpf_prog *prog;
	int ret;

	if (CHECK_ATTR(BPF_LINK_CREATE))
		return -EINVAL;

	ptype = attach_type_to_prog_type(attr->link_create.attach_type);
	if (ptype == BPF_PROG_TYPE_UNSPEC)
		return -EINVAL;

	prog = bpf_prog_get_type(attr->link_create.prog_fd, ptype);
	if (IS_ERR(prog))
		return PTR_ERR(prog);

	ret = (*klpe_bpf_prog_attach_check_attach_type)(prog,
						attr->link_create.attach_type);
	if (ret)
		goto err_out;

	switch (ptype) {
	case BPF_PROG_TYPE_CGROUP_SKB:
	case BPF_PROG_TYPE_CGROUP_SOCK:
	case BPF_PROG_TYPE_CGROUP_SOCK_ADDR:
	case BPF_PROG_TYPE_SOCK_OPS:
	case BPF_PROG_TYPE_CGROUP_DEVICE:
	case BPF_PROG_TYPE_CGROUP_SYSCTL:
	case BPF_PROG_TYPE_CGROUP_SOCKOPT:
		ret = (*klpe_cgroup_bpf_link_attach)(attr, prog);
		break;
	case BPF_PROG_TYPE_TRACING:
		ret = klpr_tracing_bpf_link_attach(attr, prog);
		break;
	case BPF_PROG_TYPE_FLOW_DISSECTOR:
	case BPF_PROG_TYPE_SK_LOOKUP:
		ret = (*klpe_netns_bpf_link_create)(attr, prog);
		break;
#ifdef CONFIG_NET
	case BPF_PROG_TYPE_XDP:
		ret = (*klpe_bpf_xdp_link_attach)(attr, prog);
		break;
#else
#error "klp-ccp: a preceeding branch should have been taken"
#endif
	default:
		ret = -EINVAL;
	}

err_out:
	if (ret < 0)
		bpf_prog_put(prog);
	return ret;
}

#define BPF_LINK_UPDATE_LAST_FIELD link_update.old_prog_fd

static int klpr_link_update(union bpf_attr *attr)
{
	struct bpf_prog *old_prog = NULL, *new_prog;
	struct bpf_link *link;
	u32 flags;
	int ret;

	if (CHECK_ATTR(BPF_LINK_UPDATE))
		return -EINVAL;

	flags = attr->link_update.flags;
	if (flags & ~BPF_F_REPLACE)
		return -EINVAL;

	link = (*klpe_bpf_link_get_from_fd)(attr->link_update.link_fd);
	if (IS_ERR(link))
		return PTR_ERR(link);

	new_prog = (*klpe_bpf_prog_get)(attr->link_update.new_prog_fd);
	if (IS_ERR(new_prog)) {
		ret = PTR_ERR(new_prog);
		goto out_put_link;
	}

	if (flags & BPF_F_REPLACE) {
		old_prog = (*klpe_bpf_prog_get)(attr->link_update.old_prog_fd);
		if (IS_ERR(old_prog)) {
			ret = PTR_ERR(old_prog);
			old_prog = NULL;
			goto out_put_progs;
		}
	} else if (attr->link_update.old_prog_fd) {
		ret = -EINVAL;
		goto out_put_progs;
	}

	if (link->ops->update_prog)
		ret = link->ops->update_prog(link, new_prog, old_prog);
	else
		ret = -EINVAL;

out_put_progs:
	if (old_prog)
		bpf_prog_put(old_prog);
	if (ret)
		bpf_prog_put(new_prog);
out_put_link:
	(*klpe_bpf_link_put)(link);
	return ret;
}

#define BPF_LINK_DETACH_LAST_FIELD link_detach.link_fd

static int klpr_link_detach(union bpf_attr *attr)
{
	struct bpf_link *link;
	int ret;

	if (CHECK_ATTR(BPF_LINK_DETACH))
		return -EINVAL;

	link = (*klpe_bpf_link_get_from_fd)(attr->link_detach.link_fd);
	if (IS_ERR(link))
		return PTR_ERR(link);

	if (link->ops->detach)
		ret = link->ops->detach(link);
	else
		ret = -EOPNOTSUPP;

	(*klpe_bpf_link_put)(link);
	return ret;
}

static int bpf_link_inc_not_zero(struct bpf_link *link)
{
	return atomic64_fetch_add_unless(&link->refcnt, 1, 0) ? 0 : -ENOENT;
}

#define BPF_LINK_GET_FD_BY_ID_LAST_FIELD link_id

static int klpr_bpf_link_get_fd_by_id(const union bpf_attr *attr)
{
	struct bpf_link *link;
	u32 id = attr->link_id;
	int fd, err;

	if (CHECK_ATTR(BPF_LINK_GET_FD_BY_ID))
		return -EINVAL;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	spin_lock_bh(&(*klpe_link_idr_lock));
	link = idr_find(&(*klpe_link_idr), id);
	/* before link is "settled", ID is 0, pretend it doesn't exist yet */
	if (link) {
		if (link->id)
			err = bpf_link_inc_not_zero(link);
		else
			err = -EAGAIN;
	} else {
		err = -ENOENT;
	}
	spin_unlock_bh(&(*klpe_link_idr_lock));

	if (err)
		return err;

	fd = (*klpe_bpf_link_new_fd)(link);
	if (fd < 0)
		(*klpe_bpf_link_put)(link);

	return fd;
}

static const struct file_operations (*klpe_bpf_stats_fops);

static int klpr_bpf_enable_runtime_stats(void)
{
	int fd;

	mutex_lock(&(*klpe_bpf_stats_enabled_mutex));

	/* Set a very high limit to avoid overflow */
	if (static_key_count(&bpf_stats_enabled_key.key) > INT_MAX / 2) {
		mutex_unlock(&(*klpe_bpf_stats_enabled_mutex));
		return -EBUSY;
	}

	fd = anon_inode_getfd("bpf-stats", &(*klpe_bpf_stats_fops), NULL, O_CLOEXEC);
	if (fd >= 0)
		static_key_slow_inc(&bpf_stats_enabled_key.key);

	mutex_unlock(&(*klpe_bpf_stats_enabled_mutex));
	return fd;
}

#define BPF_ENABLE_STATS_LAST_FIELD enable_stats.type

static int klpr_bpf_enable_stats(union bpf_attr *attr)
{

	if (CHECK_ATTR(BPF_ENABLE_STATS))
		return -EINVAL;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	switch (attr->enable_stats.type) {
	case BPF_STATS_RUN_TIME:
		return klpr_bpf_enable_runtime_stats();
	default:
		break;
	}
	return -EINVAL;
}

#define BPF_ITER_CREATE_LAST_FIELD iter_create.flags

static int klpr_bpf_iter_create(union bpf_attr *attr)
{
	struct bpf_link *link;
	int err;

	if (CHECK_ATTR(BPF_ITER_CREATE))
		return -EINVAL;

	if (attr->iter_create.flags)
		return -EINVAL;

	link = (*klpe_bpf_link_get_from_fd)(attr->iter_create.link_fd);
	if (IS_ERR(link))
		return PTR_ERR(link);

	err = (*klpe_bpf_iter_new_fd)(link);
	(*klpe_bpf_link_put)(link);

	return err;
}

__SYSCALL_DEFINEx(3, _klpp_bpf, int, cmd, union bpf_attr __user *, uattr, unsigned int, size)
{
	union bpf_attr attr;
	int err;

	if ((*klpe_sysctl_unprivileged_bpf_disabled) && !bpf_capable())
		return -EPERM;

	err = (*klpe_bpf_check_uarg_tail_zero)(uattr, sizeof(attr), size);
	if (err)
		return err;
	size = min_t(u32, size, sizeof(attr));

	/* copy attributes from user space, may be less than sizeof(bpf_attr) */
	memset(&attr, 0, sizeof(attr));
	if (copy_from_user(&attr, uattr, size) != 0)
		return -EFAULT;

	err = (*klpe_security_bpf)(cmd, &attr, size);
	if (err < 0)
		return err;

	switch (cmd) {
	case BPF_MAP_CREATE:
		err = klpr_map_create(&attr);
		break;
	case BPF_MAP_LOOKUP_ELEM:
		err = klpr_map_lookup_elem(&attr);
		break;
	case BPF_MAP_UPDATE_ELEM:
		err = klpp_map_update_elem(&attr);
		break;
	case BPF_MAP_DELETE_ELEM:
		err = klpp_map_delete_elem(&attr);
		break;
	case BPF_MAP_GET_NEXT_KEY:
		err = klpr_map_get_next_key(&attr);
		break;
	case BPF_MAP_FREEZE:
		err = klpr_map_freeze(&attr);
		break;
	case BPF_PROG_LOAD:
		err = (*klpe_bpf_prog_load)(&attr, uattr);
		break;
	case BPF_OBJ_PIN:
		err = klpr_bpf_obj_pin(&attr);
		break;
	case BPF_OBJ_GET:
		err = klpr_bpf_obj_get(&attr);
		break;
	case BPF_PROG_ATTACH:
		err = klpr_bpf_prog_attach(&attr);
		break;
	case BPF_PROG_DETACH:
		err = klpr_bpf_prog_detach(&attr);
		break;
	case BPF_PROG_QUERY:
		err = klpr_bpf_prog_query(&attr, uattr);
		break;
	case BPF_PROG_TEST_RUN:
		err = klpr_bpf_prog_test_run(&attr, uattr);
		break;
	case BPF_PROG_GET_NEXT_ID:
		err = (*klpe_bpf_obj_get_next_id)(&attr, uattr,
					  &(*klpe_prog_idr), &(*klpe_prog_idr_lock));
		break;
	case BPF_MAP_GET_NEXT_ID:
		err = (*klpe_bpf_obj_get_next_id)(&attr, uattr,
					  &(*klpe_map_idr), &(*klpe_map_idr_lock));
		break;
	case BPF_BTF_GET_NEXT_ID:
		err = (*klpe_bpf_obj_get_next_id)(&attr, uattr,
					  &(*klpe_btf_idr), &(*klpe_btf_idr_lock));
		break;
	case BPF_PROG_GET_FD_BY_ID:
		err = klpr_bpf_prog_get_fd_by_id(&attr);
		break;
	case BPF_MAP_GET_FD_BY_ID:
		err = klpr_bpf_map_get_fd_by_id(&attr);
		break;
	case BPF_OBJ_GET_INFO_BY_FD:
		err = (*klpe_bpf_obj_get_info_by_fd)(&attr, uattr);
		break;
	case BPF_RAW_TRACEPOINT_OPEN:
		err = (*klpe_bpf_raw_tracepoint_open)(&attr);
		break;
	case BPF_BTF_LOAD:
		err = klpr_bpf_btf_load(&attr);
		break;
	case BPF_BTF_GET_FD_BY_ID:
		err = klpr_bpf_btf_get_fd_by_id(&attr);
		break;
	case BPF_TASK_FD_QUERY:
		err = klpr_bpf_task_fd_query(&attr, uattr);
		break;
	case BPF_MAP_LOOKUP_AND_DELETE_ELEM:
		err = klpp_map_lookup_and_delete_elem(&attr);
		break;
	case BPF_MAP_LOOKUP_BATCH:
		err = klpp_bpf_map_do_batch(&attr, uattr, BPF_MAP_LOOKUP_BATCH);
		break;
	case BPF_MAP_LOOKUP_AND_DELETE_BATCH:
		err = klpp_bpf_map_do_batch(&attr, uattr,
				       BPF_MAP_LOOKUP_AND_DELETE_BATCH);
		break;
	case BPF_MAP_UPDATE_BATCH:
		err = klpp_bpf_map_do_batch(&attr, uattr, BPF_MAP_UPDATE_BATCH);
		break;
	case BPF_MAP_DELETE_BATCH:
		err = klpp_bpf_map_do_batch(&attr, uattr, BPF_MAP_DELETE_BATCH);
		break;
	case BPF_LINK_CREATE:
		err = klpr_link_create(&attr);
		break;
	case BPF_LINK_UPDATE:
		err = klpr_link_update(&attr);
		break;
	case BPF_LINK_GET_FD_BY_ID:
		err = klpr_bpf_link_get_fd_by_id(&attr);
		break;
	case BPF_LINK_GET_NEXT_ID:
		err = (*klpe_bpf_obj_get_next_id)(&attr, uattr,
					  &(*klpe_link_idr), &(*klpe_link_idr_lock));
		break;
	case BPF_ENABLE_STATS:
		err = klpr_bpf_enable_stats(&attr);
		break;
	case BPF_ITER_CREATE:
		err = klpr_bpf_iter_create(&attr);
		break;
	case BPF_LINK_DETACH:
		err = klpr_link_detach(&attr);
		break;
	default:
		err = -EINVAL;
		break;
	}

	return err;
}



#include <linux/kernel.h>
#include <linux/module.h>
#include "livepatch_bsc1192991.h"
#include "../kallsyms_relocs.h"

static struct klp_kallsyms_reloc klp_funcs[] = {
	{ "__bpf_map_get", (void *)&klpe___bpf_map_get },
	{ "__bpf_map_inc_not_zero", (void *)&klpe___bpf_map_inc_not_zero },
	{ "bpf_check_uarg_tail_zero", (void *)&klpe_bpf_check_uarg_tail_zero },
	{ "bpf_fd_array_map_update_elem",
	  (void *)&klpe_bpf_fd_array_map_update_elem },
	{ "bpf_fd_htab_map_update_elem",
	  (void *)&klpe_bpf_fd_htab_map_update_elem },
	{ "bpf_fd_reuseport_array_update_elem",
	  (void *)&klpe_bpf_fd_reuseport_array_update_elem },
	{ "bpf_get_file_flag", (void *)&klpe_bpf_get_file_flag },
	{ "bpf_get_perf_event_info", (void *)&klpe_bpf_get_perf_event_info },
	{ "bpf_iter_link_attach", (void *)&klpe_bpf_iter_link_attach },
	{ "bpf_iter_new_fd", (void *)&klpe_bpf_iter_new_fd },
	{ "bpf_link_fops", (void *)&klpe_bpf_link_fops },
	{ "bpf_link_get_from_fd", (void *)&klpe_bpf_link_get_from_fd },
	{ "bpf_link_new_fd", (void *)&klpe_bpf_link_new_fd },
	{ "bpf_link_put", (void *)&klpe_bpf_link_put },
	{ "bpf_map_charge_finish", (void *)&klpe_bpf_map_charge_finish },
	{ "bpf_map_charge_move", (void *)&klpe_bpf_map_charge_move },
	{ "bpf_map_copy_value", (void *)&klpe_bpf_map_copy_value },
	{ "bpf_map_new_fd", (void *)&klpe_bpf_map_new_fd },
	{ "bpf_map_offload_delete_elem",
	  (void *)&klpe_bpf_map_offload_delete_elem },
	{ "bpf_map_offload_get_next_key",
	  (void *)&klpe_bpf_map_offload_get_next_key },
	{ "bpf_map_offload_ops", (void *)&klpe_bpf_map_offload_ops },
	{ "bpf_map_offload_update_elem",
	  (void *)&klpe_bpf_map_offload_update_elem },
	{ "bpf_map_put_with_uref", (void *)&klpe_bpf_map_put_with_uref },
	{ "bpf_map_types", (void *)&klpe_bpf_map_types },
	{ "bpf_obj_get_info_by_fd", (void *)&klpe_bpf_obj_get_info_by_fd },
	{ "bpf_obj_get_next_id", (void *)&klpe_bpf_obj_get_next_id },
	{ "bpf_obj_get_user", (void *)&klpe_bpf_obj_get_user },
	{ "bpf_obj_name_cpy", (void *)&klpe_bpf_obj_name_cpy },
	{ "bpf_obj_pin_user", (void *)&klpe_bpf_obj_pin_user },
	{ "bpf_percpu_array_update", (void *)&klpe_bpf_percpu_array_update },
	{ "bpf_percpu_cgroup_storage_update",
	  (void *)&klpe_bpf_percpu_cgroup_storage_update },
	{ "bpf_percpu_hash_update", (void *)&klpe_bpf_percpu_hash_update },
	{ "bpf_prog_active", (void *)&klpe_bpf_prog_active },
	{ "bpf_prog_attach_check_attach_type",
	  (void *)&klpe_bpf_prog_attach_check_attach_type },
	{ "bpf_prog_by_id", (void *)&klpe_bpf_prog_by_id },
	{ "bpf_prog_get", (void *)&klpe_bpf_prog_get },
	{ "bpf_prog_load", (void *)&klpe_bpf_prog_load },
	{ "bpf_prog_new_fd", (void *)&klpe_bpf_prog_new_fd },
	{ "bpf_raw_tp_link_lops", (void *)&klpe_bpf_raw_tp_link_lops },
	{ "bpf_raw_tracepoint_open", (void *)&klpe_bpf_raw_tracepoint_open },
	{ "bpf_stats_enabled_mutex", (void *)&klpe_bpf_stats_enabled_mutex },
	{ "bpf_stats_fops", (void *)&klpe_bpf_stats_fops },
	{ "bpf_xdp_link_attach", (void *)&klpe_bpf_xdp_link_attach },
	{ "btf_find_spin_lock", (void *)&klpe_btf_find_spin_lock },
	{ "btf_get_by_fd", (void *)&klpe_btf_get_by_fd },
	{ "btf_get_fd_by_id", (void *)&klpe_btf_get_fd_by_id },
	{ "btf_idr", (void *)&klpe_btf_idr },
	{ "btf_idr_lock", (void *)&klpe_btf_idr_lock },
	{ "btf_new_fd", (void *)&klpe_btf_new_fd },
	{ "btf_put", (void *)&klpe_btf_put },
	{ "btf_type_by_id", (void *)&klpe_btf_type_by_id },
	{ "btf_type_id_size", (void *)&klpe_btf_type_id_size },
	{ "cgroup_bpf_link_attach", (void *)&klpe_cgroup_bpf_link_attach },
	{ "cgroup_bpf_prog_attach", (void *)&klpe_cgroup_bpf_prog_attach },
	{ "cgroup_bpf_prog_detach", (void *)&klpe_cgroup_bpf_prog_detach },
	{ "cgroup_bpf_prog_query", (void *)&klpe_cgroup_bpf_prog_query },
	{ "get_files_struct", (void *)&klpe_get_files_struct },
	{ "link_idr", (void *)&klpe_link_idr },
	{ "link_idr_lock", (void *)&klpe_link_idr_lock },
	{ "map_idr", (void *)&klpe_map_idr },
	{ "map_idr_lock", (void *)&klpe_map_idr_lock },
	{ "netns_bpf_link_create", (void *)&klpe_netns_bpf_link_create },
	{ "netns_bpf_prog_attach", (void *)&klpe_netns_bpf_prog_attach },
	{ "netns_bpf_prog_detach", (void *)&klpe_netns_bpf_prog_detach },
	{ "netns_bpf_prog_query", (void *)&klpe_netns_bpf_prog_query },
	{ "perf_get_event", (void *)&klpe_perf_get_event },
	{ "prog_idr", (void *)&klpe_prog_idr },
	{ "prog_idr_lock", (void *)&klpe_prog_idr_lock },
	{ "put_files_struct", (void *)&klpe_put_files_struct },
	{ "security_bpf", (void *)&klpe_security_bpf },
	{ "security_bpf_map_alloc", (void *)&klpe_security_bpf_map_alloc },
	{ "security_bpf_map_free", (void *)&klpe_security_bpf_map_free },
	{ "sock_map_get_from_fd", (void *)&klpe_sock_map_get_from_fd },
	{ "sock_map_prog_detach", (void *)&klpe_sock_map_prog_detach },
	{ "sock_map_update_elem_sys", (void *)&klpe_sock_map_update_elem_sys },
	{ "sysctl_unprivileged_bpf_disabled",
	  (void *)&klpe_sysctl_unprivileged_bpf_disabled },
};

int livepatch_bsc1192991_init(void)
{
	return __klp_resolve_kallsyms_relocs(klp_funcs, ARRAY_SIZE(klp_funcs));
}
